"""
"""


#[
from __future__ import annotations

import json as _js
import copy as _co
import types as _ty
import numpy as _np
import re as _re
import operator as _op
import functools as _ft
from typing import (Self, TypeAlias, Literal, Sequence, Protocol, Any, )
from collections.abc import (Iterable, Callable, )
from numbers import (Number, )

from ..conveniences import views as _views
from ..conveniences import descriptions as _descriptions
from ..series import main as _series
from .. import dataslabs as _dataslabs
from .. import dates as _dates

from . import _imports as _imports
from . import _exports as _exports
#]


__all__ = [
    "Databank",
]


SourceNames: TypeAlias = Iterable[str] | str | Callable[[str], bool] | None
TargetNames: TypeAlias = Iterable[str] | str | Callable[[str], str] | None
InterpretRange: TypeAlias = Literal["base"] | Literal["extended"]


class SteadyDatabankableProtocol(Protocol):
    """
    """
    #[
    def get_min_max_shifts(self, *args) -> tuple(int, int): ...
    def _get_steady_databank(self, *args) -> Any: ...
    #]


class EmptyDateRange(Exception):
    def __init__(self):
        super().__init__("Empty date range is not allowed in this context.")


def _extended_range_tuple_from_base_range(input_range, min_shift, max_shift):
    """
    """
    range_list = [t for t in input_range]
    start_date, end_date = range_list[0], range_list[-1]
    start_date += min_shift
    end_date += max_shift
    return start_date, end_date


def _extended_range_tuple_from_extended_range(input_range, min_shift, max_shift):
    """
    """
    range_list = [t for t in input_range]
    start_date, end_date = range_list[0], range_list[-1]
    return start_date, end_date


_EXTENDED_RANGE_TUPLE_RESOLUTION = {
    "base": _extended_range_tuple_from_base_range,
    "extended": _extended_range_tuple_from_extended_range,
}


_SERIES_CONSTRUCTOR_RESOLUTION = {
    "start_date": _series.Series.from_start_date_and_values,
    "range": _series.Series.from_dates_and_values,
}


_ARRAY_TRANSPOSER_RESOLUTION = {
    "vertical": _np.transpose,
    "horizontal": lambda x: x,
}


class Databank(
    _imports.DatabankImportMixin,
    _exports.DatabankExportMixin,
    _descriptions.DescriptionMixin,
    _views.DatabankViewMixin,
    dict,
):
    """
    Create a databank object as a simple namespace with utility functions
    """
    #[
    def __init__(
        self,
        /,
        description: str = "",
    ) -> None:
        self.set_description(description)

    @classmethod
    def from_dict(
        cls,
        _dict: dict,
        /,
    ) -> Self:
        """
        """
        self = cls()
        for k, v in _dict.items():
            self[k] = v
        return self

    @classmethod
    def from_array(
        cls,
        array: _np.ndarray,
        qid_to_name: Sequence[str] | dict[int, str],
        dates: _dates.Dater,
        /,
        add_to_databank: Self | None = None,
        qid_to_description: dict[int, str] | None = None,
        array_orientation: Literal["vertical"] | Literal["horizontal"] = "vertical",
        interpret_dates: Literal["start_date"] | Literal["range"] = "start_date",
    ) -> Self:
        """
        """
        self = add_to_databank if add_to_databank else cls()
        constructor = _SERIES_CONSTRUCTOR_RESOLUTION[interpret_dates]
        transposer = _ARRAY_TRANSPOSER_RESOLUTION[array_orientation]
        for qid, data in enumerate(transposer(array)):
            name = qid_to_name.get(qid, None)
            if not name:
                continue
            description = qid_to_description[qid] if qid_to_description else ""
            series = constructor(dates, data.reshape(-1, 1), description=description)
            self[name] = series
        return self

    @classmethod
    def for_simulation(
        cls,
        simulatable: SimulatableProtocol,
        base_range: Iterable[_dates.Dater],
        /,
        value: Number | _np.ndarray | None = None,
        func: Callable | None = None
    ) -> Self:
        """
        """
        ext_range, *_ = _dataslabs.get_extended_range(simulatable, base_range, )
        if value is None and func is None:
            value = 0
        if value is not None:
            constructor = _ft.partial(_series.Series.from_dates_and_values, ext_range, value, )
        else:
            constructor = _ft.partial(_series.Series.from_dates_and_func, ext_range, func, )
        self = cls()
        for name in simulatable.get_databank_names():
            self[name] = constructor()
        return self

    def get_names(self: Self) -> list[str]:
        """
        Get all names stored in a databank save for private attributes
        """
        return list(self.keys())

    def get_num_items(self: Self) -> int:
        """
        """
        return len(self.keys())

    def to_dict(self: Self) -> dict:
        """
        Convert Databank to dict
        """
        return { k: v for k, v in self.items() }

    def copy(
        self: Self,
        /,
        source_names: SourceNames = None,
        target_names: TargetNames = None,
    ) -> Self:
        """
        """
        new_databank = _co.deepcopy(self)
        if source_names is None and target_names is None:
            return new_databank
        context_names = new_databank.get_names()
        source_names, target_names = _resolve_source_target_names(source_names, target_names, context_names)
        new_databank = new_databank.rename(source_names, target_names)
        new_databank.keep(target_names)
        return new_databank

    def rename(
        self: Self,
        /,
        source_names: SourceNames = None,
        target_names: TargetNames = None,
    ) -> Self:
        """
        """
        context_names = self.get_names()
        source_names, target_names = _resolve_source_target_names(source_names, target_names, context_names)
        for old_name, new_name in (
            z for z in zip(source_names, target_names) if z[0] in context_names and z[0]!=z[1]
        ):
            self[new_name] = self.pop(old_name)
        return self

    def remove(
        self: Self,
        /,
        remove_names: SourceNames = None,
    ) -> Self:
        """
        """
        if remove_names is None:
            return self
        context_names = self.get_names()
        remove_names, *_ = _resolve_source_target_names(remove_names, None, context_names)
        for n in set(remove_names).intersection(context_names):
            del self[n]
        return self

    def keep(
        self: Self,
        /,
        keep_names: SourceNames = None,
    ) -> Self:
        """
        """
        if keep_names is None:
            return self
        context_names = self.get_names()
        keep_names, *_ = _resolve_source_target_names(keep_names, None, context_names)
        remove_names = set(context_names).difference(keep_names)
        return self.remove(remove_names)

    def apply(
        self,
        func: Callable,
        /,
        source_names: SourceNames = None,
        target_names: TargetNames = None,
    ) -> Self:
        context_names = self.get_names()
        source_names, target_names = _resolve_source_target_names(source_names, target_names, context_names)
        for s, t in zip(source_names, target_names):
            self[t] = func(self[s])
        return self

    def filter(
        self,
        name_test: Callable | None = None,
        value_test: Callable | None = None,
    ) -> Iterable[str]:
        """
        """
        names = self.get_names()
        if name_test is None and value_test is None:
            return names
        name_test = name_test if name_test else lambda x: True
        value_test = value_test if value_test else lambda x: True
        return [ n for n in names if name_test(n) and value_test(self[n]) ]

    def get_series_names_by_frequency(
        self,
        frequency: _dates.Frequency,
    ) -> Iterable[str]:
        """
        """
        return self.filter(value_test=lambda x: isinstance(x, _series.Series) and x.frequency==frequency)

    def get_range_by_frequency(
        self,
        frequency: _dates.Frequency,
    ) -> Ranger:
        names = self.get_series_names_by_frequency(frequency)
        if not names:
            return Ranger(None, None)
        min_start_date = min((self[n].start_date for n in names), key=_op.attrgetter("serial"))
        max_end_date = max((self[n].end_date for n in names), key=_op.attrgetter("serial"))
        return _dates.Ranger(min_start_date, max_end_date)

    def to_json(self, **kwargs):
        return _js.dumps(self, **kwargs)

    def underlay(
        self,
        other: Self,
        /,
    ) -> None:
        """"
        """
        self_names = self.filter(value_test=lambda x: isinstance(x, _series.Series))
        other_names = other.filter(value_test=lambda x: isinstance(x, _series.Series))
        names = set(self_names).intersection(other_names)
        for n in names:
            if self[n].frequency == other[n].frequency:
                self[n].underlay(other[n], )

    def clip(
        self,
        new_start_date: _dates.Dater | None = None,
        new_end_date: _dates.Dater | None = None,
        /,
    ) -> None:
        if new_start_date is None and new_end_date is None:
            return
        frequency = new_start_date.frequency if new_start_date is not None else new_end_date.frequency
        names = self.filter(value_test=lambda x: isinstance(x, _series.Series) and x.frequency == frequency)
        for n in names:
            self[n].clip(new_start_date, new_end_date, )

    def prepend(
        self,
        prepending: Self,
        end_prepending: _dates.Dater,
        /,
    ) -> Self:
        """
        """
        prepending = prepending.copy()
        prepending.clip(None, end_prepending, )
        self.underlay(prepending, )

    def add_steady(
        self,
        steady_databankable: SteadyDatabankableProtocol,
        input_range: Iterable[_dates.Dater],
        /,
        deviation: bool = False,
        interpret_range: InterpretRange = "base",
    ) -> Self:
        """
        """
        min_shift, max_shift = steady_databankable.get_min_max_shifts()
        start_date, end_date = _resolve_input_range(input_range, min_shift, max_shift, interpret_range)
        num_columns = int(end_date - start_date + 1)
        if num_columns < 1:
            raise Exception("Empty date range is not allowed when creating steady databank")
        steady_databank = steady_databankable._get_steady_databank(start_date, end_date, deviation=deviation)
        self.update(steady_databank)

    add_zero = _ft.partialmethod(add_steady, deviation=True)

###     def __getitem__(self, name):
###         return self.__dict__[name]
### 
###     def __setitem__(self, name, value) -> None:
###         self.__dict__[name] = value
### 
###     def __or__(self, other) -> Self:
###         new = _co.deepcopy(self)
###         new.__dict__.update(other.__dict__)
###         return new
### 
###     def _update(
###         self,
###         other: Databank,
###         /,
###     ) -> Self:
###         """
###         Update self using items from other
###         """
###         self.__dict__.update(other.__dict__)
    #]


#
# Add databank methods without the leading underscore
#
### exceptions = ["_description", ]
### single_underscore_names = [
###     n for n in dir(Databank) 
###     if n.startswith("_") and not n.startswith("__") and not n.endswith("_") and n not in exceptions
### ]
### for n in single_underscore_names:
###     setattr(Databank, n.removeprefix("_"), getattr(Databank, n))


def _resolve_source_target_names(
    source_names: SourceNames,
    target_names: TargetNames,
    context_names: Iterable[str],
    /,
) -> tuple[Iterable[str], Iterable[str]]:
    """
    """
    if source_names is None:
        source_names = context_names
    if isinstance(source_names, str):
        source_names = [ source_names ]
    if callable(source_names):
        func = source_names
        source_names = [ n for n in context_names if func(n) ]
    if target_names is None:
        target_names = source_names
    if isinstance(target_names, str):
        target_names = [ target_names ]
    if callable(target_names):
        func = target_names
        target_names = [ func(n) for n in source_names ]
    return source_names, target_names


def _resolve_input_range(
    input_range: Iterable[_dates.Dater],
    min_shift: int,
    max_shift: int,
    interpret_range: InterpretRange,
) -> tuple[Dater, Dater]:
    return _EXTENDED_RANGE_TUPLE_RESOLUTION[interpret_range](input_range, min_shift, max_shift)




"""
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Callable, )
import dataclasses as _dc
import copy as _cp

from ..incidences import main as _incidence
from .. import equations as _equations
from .. import wrongdoings as _wrongdoings
from ..aldi import adaptations as _adaptations

from .. import transforms as _transforms
#]


__all__ = (
    "Explanatory",
)


@_dc.dataclass
class Explanatory:
    """
    """
    #[
    _RESIDUAL_PREFIX = "res_"
    _IDENTITY_EQUAL_SIGN = "==="
    equation: _equations.Equation | None = None
    is_identity: bool | None = None
    _lhs_human: str | None = None
    lhs_name: str | None = None
    res_name: str | None = None
    res_name: str | None = None
    lhs_qid: int | None = None
    _rhs_human: str | None = None
    _lhs_transform: TransformProtocol | None = None
    _eval_level_str: str | None = None
    _eval_res_str: str | None = None
    _custom_functions: dict[str, Callable] | None = None
    all_names: tuple(str) | None = None
    eval_level: Callable | None = None
    eval_res: Callable | None = None

    def __init__(
        self,
        equation: _equations.Equation,
        /,
        custom_functions: dict[str, Callable] | None = None,
    ) -> None:
        """
        """
        self.equation = _cp.deepcopy(equation, )
        self._detect_identity()
        self._split_equation()
        self._parse_lhs()
        self._add_residual_to_rhs()
        self._collect_all_names()
        self._custom_functions = _cp.deepcopy(custom_functions, )
        self._custom_functions = _adaptations.add_function_adaptations_to_custom_functions(custom_functions, )

    def finalize(
        self,
        name_to_qid,
        /,
    ) -> None:
        """
        """
        self.equation.finalize(name_to_qid, )
        self.lhs_qid = name_to_qid[self.lhs_name]
        self.res_qid = name_to_qid[self.res_name] if self.res_name is not None else None
        rhs_xtring, *_ = _equations.xtring_from_human(self._rhs_human, name_to_qid, )
        lhs_xtring, *_ = _equations.xtring_from_human(self._lhs_human, name_to_qid, )
        lhs_token = _incidence.Token(self.lhs_qid, 0, )
        self._create_eval_level(lhs_token, rhs_xtring, )
        self._create_eval_res(lhs_xtring, rhs_xtring, )

    @property
    def min_shift(self, /, ) -> int:
        """
        """
        return _incidence.get_min_shift(self.equation.incidence, )

    @property
    def max_shift(self, /, ) -> int:
        """
        """
        return _incidence.get_max_shift(self.equation.incidence, )


    @property
    def description(self, /, ) -> str:
        """
        """
        return (
            str(self.equation.description)
            if self.equation.description is not None else ""
        )

    def print_equation(
        self,
        /,
        indent: int = 0,
        description: bool = True,
    ) -> str:
        """
        """
        indented = " " * indent
        print_description = description and self.equation.description
        return (
            ((f"{indented}\"{self.equation.description}\"\n") if print_description else "")
            +f"{indented}{self.equation.human}"
        )

    def _detect_identity(self, /, ) -> None:
        """
        """
        self.is_identity = self._IDENTITY_EQUAL_SIGN in self.equation.human

    def _split_equation(self, ) -> None:
        """
        """
        human = self.equation.human
        equal_sign = "=" if not self.is_identity else self._IDENTITY_EQUAL_SIGN
        split = human.split(equal_sign, )
        if len(split) != 2:
            raise _wrongdoings.IrisPieError(
                f"This equation is not in a LHS=RHS or LHS{self._IDENTITY_EQUAL_SIGN}RHS form: {self.equation.human}"
            )
        self._lhs_human, self._rhs_human = split

    def _parse_lhs(self, ) -> None:
        transform, lhs_name = _transforms.recognize_transform_in_equation(self._lhs_human, )
        if transform is None:
            raise _wrongdoings.IrisPieError(f"Could not parse this LHS expression: {self.lhs_name}")
        self._lhs_transform = transform
        self.lhs_name = lhs_name

    def _add_residual_to_rhs(self, /, ) -> None:
        """
        """
        if self.is_identity:
            self.res_name = None
        else:
            self.res_name = f"{self._RESIDUAL_PREFIX}{self.lhs_name}"
            self._rhs_human += f"+{self.res_name}"
            self.equation.human += f"+{self.res_name}"

    def _collect_all_names(self, /, ) -> None:
        """
        """
        self.all_names = tuple(set(
            _equations.generate_names_from_human(self.equation.human, )
        )) + ((self.res_name, ) if self.res_name is not None else ())

    def _create_eval_level(
        self,
        lhs_token: str,
        rhs_xtring: str,
        /,
    ) -> None:
        self._eval_level_str = (
            "lambda x, t: "
            + self._lhs_transform.create_eval_level_str(lhs_token, rhs_xtring, )
        )
        self.eval_level = eval(self._eval_level_str, self._custom_functions, )

    def _create_eval_res(
        self,
        lhs_xtring: str,
        rhs_xtring: str,
        /,
    ) -> None:
        if not self.is_identity:
            self._eval_res_str = (
                "lambda x, t: "
                f"{lhs_xtring}-({rhs_xtring})"
            )
            self.eval_res = eval(self._eval_res_str, self._custom_functions, )

    def __str__(self, /, ) -> str:
        """
        """
        indented = " " * 4
        return "\n".join((
            f"{indented}Explanatory object",
            f"{indented}Description: \"{self.description}\"",
            f"{indented}Equation: {self.equation.human}",
            f"{indented}LHS name: {self.lhs_name}",
            f"{indented}LHS transform: {str(self._lhs_transform)}",
            f"{indented}Residual name: {self.res_name}",
        ))

    def precopy(self, /, ) -> None:
        """
        """

    def simulate(
        self,
        data: _np.ndarray,
        columns: int | _np.ndarray,
        /,
    ) -> None:
        """
        """
        row = self.lhs_qid
        data[row, columns] = self.eval_level(data, columns, )

    def exogenize(
        self,
        data: _np.ndarray,
        columns: int | _np.ndarray,
        values: int | _np.ndarray,
        /,
    ) -> None:
        """
        """
        lhs_row = self.lhs_qid
        res_row = self.res_qid
        data[lhs_row, columns] = values
        data[res_row, columns] = self.eval_res(data, columns, )
    #]




"""
Model quantities
"""


#[
from __future__ import annotations

import enum
import dataclasses as _dc

from typing import (TypeAlias, )
from collections.abc import (Iterable, )
#]


class QuantityKind(enum.Flag):
    """
    Classification of model quantities
    """
    #[
    UNSPECIFIED = enum.auto()
    TRANSITION_VARIABLE = enum.auto()
    TRANSITION_SHOCK = enum.auto()
    MEASUREMENT_VARIABLE = enum.auto()
    MEASUREMENT_SHOCK = enum.auto()
    EXOGENOUS_VARIABLE = enum.auto()
    LHS_VARIABLE = enum.auto()
    RHS_ONLY_VARIABLE = enum.auto()
    PARAMETER = enum.auto()
    TRANSITION_STD = enum.auto()
    MEASUREMENT_STD = enum.auto()

    ENDOGENOUS_VARIABLE = TRANSITION_VARIABLE | MEASUREMENT_VARIABLE
    VARIABLE = ENDOGENOUS_VARIABLE | EXOGENOUS_VARIABLE
    STD = TRANSITION_STD | MEASUREMENT_STD
    PARAMETER_OR_STD = PARAMETER | STD
    SHOCK = TRANSITION_SHOCK | MEASUREMENT_SHOCK
    #]


_export_kinds  = [
    "TRANSITION_VARIABLE", "TRANSITION_SHOCK", "TRANSITION_STD",
    "MEASUREMENT_VARIABLE", "MEASUREMENT_SHOCK", "MEASUREMENT_STD",
]

for n in _export_kinds:
    exec(f"{n} = QuantityKind.{n}")
__all__ = ["filter_quantities_by_name", ] + _export_kinds


@_dc.dataclass(slots=True, )
class Quantity:
    """
    """
    #[
    id: int | None = None
    human: str | None = None
    kind: QuantityKind = QuantityKind.UNSPECIFIED
    logly: bool | None = None
    description: str | None = None
    entry: int | None = None

    def set_id(self, qid: int) -> Self:
        self.id = qid
        return self

    def set_logly(self, logly: bool) -> Self:
        self.logly = logly
        return self

    def print_name_maybe_log(self, /, ) -> str:
        return print_name_maybe_log(self.human, self.logly, )

    def __hash__(self, /, ) -> int:
        return hash(self.__repr__)
    #]


def create_name_to_qid(quantities: Iterable[Quantity]) -> dict[str, int]:
    return { qty.human: qty.id for qty in quantities }


def create_name_to_quantity(quantities: Iterable[Quantity]) -> dict[str, int]:
    return { qty.human: qty for qty in quantities }


def create_qid_to_name(quantities: Iterable[Quantity]) -> dict[int, str]:
    return { qty.id: qty.human for qty in quantities }


def create_qid_to_description(quantities: Iterable[Quantity]) -> dict[int, str]:
    return { qty.id: qty.description for qty in quantities }


def create_qid_to_kind(quantities: Iterable[Quantity]) -> dict[int, str]:
    return { qty.id: qty.kind for qty in quantities }


def generate_quantities_of_kind(quantities: Iterable[Quantity], kind: QuantityKind | None) -> Iterable[Quantity]:
    is_of_kind = (lambda qty: qty.kind in kind) if kind is not None else lambda k: True
    return ( qty for qty in quantities if is_of_kind(qty) )


def generate_qids_by_kind(quantities: Iterable[Quantity], kind: QuantityKind) -> list[int]:
    return ( qty.id for qty in quantities if qty.kind in kind )


def generate_quantity_names_by_kind(quantities: Iterable[Quantity], kind: QuantityKind) -> list[str]:
    return ( qty.human for qty in quantities if qty.kind in kind )


def generate_all_quantity_names(quantities: Iterable[Quantity]) -> Iterable[str]:
    return ( qty.human for qty in quantities )


def generate_all_qids(quantities: Iterable[Quantity]) -> Iterable[int]:
    return ( qty.id for qty in quantities )


def get_max_qid(quantities: Iterable[Quantity]) -> int:
    return max(qty.id for qty in quantities)


def create_qid_to_logly(quantities: Iterable[Quantity]) -> dict[int, bool]:
    return { qty.id: qty.logly for qty in quantities }


def change_logly(
    quantities: Iterable[Quantity],
    new_logly: bool,
    qids: Iterable[int],
    /
) -> Iterable[Quantity]:
    qids = set(qids)
    return [
        qty if qty.id not in qids or qty.logly is None else Quantity(qty.id, qty.human, qty.kind, new_logly)
        for qty in quantities
    ]


def validate_selection_of_quantities(
    allowed_quantities: Iterable[Quantity],
    custom_quantities: Iterable[Quantity] | None,
    /,
) -> tuple[Iterable[Quantity], Iterable[Quantity]]:
    """
    """
    invalid_quantities = list(set(custom_quantities) - set(allowed_quantities)) if custom_quantities is not None else []
    custom_quantities = list(custom_quantities) if custom_quantities is not None else list(allowed_quantities)
    return custom_quantities, invalid_quantities


def lookup_quantities_by_name(
    quantities: Iterable[Quantity],
    custom_names: Iterable[str],
    /,
) -> tuple[Iterable[Quantity], tuple[str]]:
    """
    Lookup quantities by name, and return a list of quantities and a list
    of invalid names
    """
    custom_names = list(custom_names)
    name_to_quantity  = create_name_to_quantity(quantities)
    custom_quantities = tuple(
        name_to_quantity[n]
        for n in custom_names if n in name_to_quantity
    )
    invalid_names = tuple(
        n for n in custom_names if n not in name_to_quantity
    )
    return custom_quantities, invalid_names


def lookup_qids_by_name(
    quantities: Iterable[Quantity],
    custom_names: Iterable[str],
    /,
) -> tuple[tuple[int], tuple[str]]:
    """
    Lookup quantities by name, and return a list of quantities and a list
    of invalid names
    """
    custom_names = list(custom_names)
    name_to_qid  = create_name_to_qid(quantities)
    custom_qids = tuple(
        name_to_qid[n]
        for n in custom_names if n in name_to_qid
    )
    invalid_names = tuple(
        n for n in custom_names if n not in name_to_qid
    )
    return custom_qids, invalid_names


def filter_quantities_by_name(
    quantities: Iterable[Quantity],
    /,
    include_names: Iterable[str] | None = None,
    exclude_names: Iterable[str] | None = None,
) -> Iterable[Quantity]:
    """
    """
    include_names = set(include_names) if include_names is not None else None
    exclude_names = set(exclude_names) if exclude_names is not None else None
    inclusion_test = lambda name: include_names is None or name in include_names
    exclusion_test = lambda name: exclude_names is None or name not in exclude_names
    return [ qty for qty in quantities if inclusion_test(qty.human) and exclusion_test(qty.human) ]


def generate_index_logly(
    qids: Iterable[int],
    qid_to_logly: dict[int, bool],
    /,
) -> Iterable[int]:
    """
    """
    return (i for i, qid in enumerate(qids, ) if qid_to_logly[qid])


def print_name_maybe_log(name, logly, /, ) -> str:
    return f"log({name})" if logly else name





"""
"""


#[
from __future__ import annotations

from typing import (Self, TypeAlias, Literal, )
from collections.abc import (Iterable, Callable, )
from numbers import (Number, )
import copy as _co
import numpy as _np
import itertools as _it
import functools as _ft

from .. import equations as _equations
from .. import quantities as _quantities
from .. import wrongdoings as _wd
from .. import sources as _sources
from .. import dates as _dates
from ..parsers import common as _pc
from ..databanks import main as _databanks
from ..fords import solutions as _sl
from ..fords import steadiers as _fs
from ..fords import descriptors as _de
from ..fords import systems as _sy

from . import variants as _variants
from . import invariants as _invariants
from . import _flags
from . import _simulate
from . import _steady
from . import _get
#]


#[
__all__ = [
    "Model"
]


_SteadySolverReturn: TypeAlias = tuple[
    _np.ndarray|None, Iterable[int]|None,
    _np.ndarray|None, Iterable[int]|None,
]


_EquationSwitch: TypeAlias = Literal["dynamic"] | Literal["steady"]
#]


class Model(
    _sources.SourceMixin,
    _simulate.SimulateMixin,
    _steady.SteadyMixin,
    _get.GetMixin,
):
    """

    """
    #[
    __slots__ = (
        "_invariant", "_variants",
    )

    def __init__(
        self,
        /,
    ) -> None:
        """
        """
        self._invariant = None
        self._variants = None

    def assign(
        self,
        /,
        **kwargs,
    ) -> Self:
        """
        """
        garbage_key = None
        qid_to_value = _rekey_dict(kwargs, _quantities.create_name_to_qid(self._invariant._quantities))
        for v in self._variants:
            v.update_values_from_dict(qid_to_value)
        #
        self._enforce_auto_values()
        return self

    def assign_from_databank(
        self,
        databank: _databanks.Databank,
        /,
    ) -> Self:
        """
        """
        return self.assign(**databank.__dict__)

    def copy(self) -> Self:
        """
        Create a deep copy of the Model object
        """
        return _co.deepcopy(self)

    def __getitem__(self, variants):
        new = self.from_self()
        index_variants = resolve_variant(self, variants)
        new._variants = [ self._variants[i] for i in index_variants ]
        return new

    def alter_num_variants(
        self,
        new_num: int,
        /,
    ) -> Self:
        """
        Alter (expand, shrink) the number of alternative parameter variants in this model object
        """
        if new_num < self.num_variants:
            self._shrink_num_variants(new_num, )
        elif new_num > self.num_variants:
            self._expand_num_variants(new_num, )
        return self

    def change_logly(
        self,
        new_logly: bool,
        some_names: Iterable[str] | None = None,
        /
    ) -> None:
        """
        Change the log-status of some Model quantities
        """
        some_names = set(some_names) if some_names else None
        qids = [
            qty.id
            for qty in self._invariant._quantities
            if qty.logly is not None and (some_names is None or qty.human in some_names)
        ]
        self._invariant._quantities = _quantities.change_logly(self._invariant._quantities, new_logly, qids)

    @property
    def num_variants(self, /, ) -> int:
        """
        Number of alternative variants within this Model
        """
        return len(self._variants)

    @property
    def is_linear(self, /, ) -> bool:
        """
        True for Models declared as linear
        """
        return self._invariant._flags.is_linear

    @property
    def is_flat(self, /, ) -> bool:
        """
        True for Models declared as flat
        """
        return self._invariant._flags.is_flat

    def create_name_to_qid(self, /, ) -> dict[str, int]:
        return _quantities.create_name_to_qid(self._invariant._quantities)

    def create_qid_to_name(self, /, ) -> dict[int, str]:
        return _quantities.create_qid_to_name(self._invariant._quantities)

    def create_qid_to_kind(self, /, ) -> dict[int, str]:
        return _quantities.create_qid_to_kind(self._invariant._quantities)

    def create_qid_to_descriptor(self, /, ) -> dict[int, str]:
        """
        Create a dictionary mapping from quantity id to quantity descriptor
        """
        return _quantities.create_qid_to_descriptor(self._invariant._quantities)

    def create_qid_to_logly(self, /, ) -> dict[int, bool]:
        """
        Create a dictionary mapping from quantity id to quantity log-status
        """
        return _quantities.create_qid_to_logly(self._invariant._quantities)

    def create_steady_array(
        self,
        /,
        variant: _variants.Variant|None = None,
        **kwargs,
    ) -> _np.ndarray:
        qid_to_logly = self.create_qid_to_logly()
        if variant is None:
            variant = self._variants[0]
        return variant.create_steady_array(qid_to_logly, **kwargs, )

    def create_zero_array(
        self,
        /,
        variant: _variants.Variant|None = None,
        **kwargs,
    ) -> _np.ndarray:
        """
        """
        qid_to_logly = self.create_qid_to_logly()
        if variant is None:
            variant = self._variants[0]
        return variant.create_zero_array(qid_to_logly, **kwargs, )

    def create_some_array(
        self,
        /,
        deviation: bool,
        **kwargs,
    ) -> _np.ndarray:
        return {
            True: self.create_zero_array, False: self.create_steady_array,
        }[deviation](**kwargs)

    def _enforce_auto_values(self, /, ) -> None:
        """
        """
        #
        # Reset levels of shocks to zero, remove changes
        #
        assign_shocks = {
            qid: (0, _np.nan)
            for qid in _quantities.generate_qids_by_kind(self._invariant._quantities, _quantities.QuantityKind.SHOCK)
        }
        self._variants[0].update_values_from_dict(assign_shocks)
        #
        # Remove changes from quantities that are not logly variables
        #
        assign_non_logly = {
            qid: (..., _np.nan)
            for qid in  _quantities.generate_qids_by_kind(self._invariant._quantities, ~sources.LOGLY_VARIABLE)
        }
        self._variants[0].update_values_from_dict(assign_non_logly)

    def _shrink_num_variants(self, new_num: int, /, ) -> None:
        """
        """
        if new_num<1:
            raise Exception('Number of variants must be one or more')
        self._variants = self._variants[0:new_num]

    def _expand_num_variants(self, new_num: int, /, ) -> None:
        """
        """
        for i in range(self.num_variants, new_num):
            self._variants.append(_co.deepcopy(self._variants[-1]))

    def systemize(
        self,
        /,
        **kwargs,
    ) -> Iterable[_sy.System]:
        """
        Create unsolved first-order system for each variant
        """
        model_flags = self._invariant._flags.update_from_kwargs(**kwargs, )
        return [
            self._systemize(variant, self._invariant._dynamic_descriptor, model_flags, )
            for variant in self._variants
        ]

    def _systemize(
        self,
        variant: _variants.Variant,
        descriptor: _de.Descriptor,
        model_flags: flags.Flags,
        /,
    ) -> _sy.System:
        """
        Create unsolved first-order system for one variant
        """
        ac = descriptor.aldi_context
        num_columns = ac.shape_data[1]
        qid_to_logly = self.create_qid_to_logly()
        #
        if model_flags.is_linear:
            data_array = variant.create_zero_array(qid_to_logly, num_columns=num_columns, shift_in_first_column=ac.min_shift)
            steady_array = variant.create_steady_array(qid_to_logly, num_columns=1, ).reshape(-1)
        else:
            data_array = variant.create_steady_array(qid_to_logly, num_columns=num_columns, )
            steady_array = data_array[:, -ac.min_shift]
        #
        return _sy.System(descriptor, data_array, steady_array, )

    def solve(
        self,
        /,
        **kwargs,
    ) -> None:
        """
        Calculate first-order solution for each Variant within this Model
        """
        model_flags = self._invariant._flags.update_from_kwargs(**kwargs, )
        for variant in self._variants:
            self._solve(variant, model_flags, )

    def _solve(
        self,
        variant: _variants.Variant,
        model_flags: flags.Flags,
        /,
    ) -> None:
        """
        Calculate first-order solution for one Variant of this Model
        """
        system = self._systemize(variant, self._invariant._dynamic_descriptor, model_flags, )
        variant.solution = _sl.Solution(self._invariant._dynamic_descriptor, system, )

    def _choose_plain_equator(
        self,
        equation_switch: _EquationSwitch,
        /,
    ) -> Callable | None:
        """
        """
        match equation_switch:
            case "dynamic":
                return self._invariant._plain_equator_for_dynamic_equations
            case "steady":
                return self._invariant._plain_equator_for_steady_equations

    def _assign_default_stds(
        self,
        default_std: Number | None = None,
        /,
    ) -> None:
        """
        """
        default_std = default_std if default_std is not None else _DEFAULT_STD_RESOLUTION(self.get_flags())
        self.assign(**{ k: default_std for k in _quantities.generate_quantity_names_by_kind(self._invariant._quantities, _quantities.QuantityKind.STD) })

    @classmethod
    def from_source(
        cls,
        source: _sources.AlgebraicSource,
        /,
        default_std: int | None = None,
        context: dict | None = None,
        **kwargs,
    ) -> Self:
        """
        """
        self = cls()
        #
        self._invariant = _invariants.Invariant(
            source,
            context=context,
            **kwargs,
        )
        #
        self._variants = [ _variants.Variant(self._invariant._quantities) ]
        #
        self._enforce_auto_values()
        self._assign_default_stds(default_std)
        #
        return self

    def from_self(self, ) -> Self:
        """
        Create a new Model object with pointers to invariant and variants of this Model object
        """
        new = type(self)()
        new._invariant = self._invariant
        new._variants = self._variants
        return new

    #
    # Implement SimulatableProtocol
    #

    def get_min_max_shifts(self) -> tuple[int, int]:
        """
        """
        return self._invariant._min_shift, self._invariant._max_shift

    def get_databank_names(self, /, ) -> list[str]:
        qid_to_name = self.create_qid_to_name()
        return [ qid_to_name[qid] for qid in range(len(qid_to_name)) ]
    #]


_DEFAULT_STD_LINEAR = 1
_DEFAULT_STD_NONLINEAR = 0.01
_DEFAULT_STD_RESOLUTION = lambda flags: {
    True: _DEFAULT_STD_LINEAR,
    False: _DEFAULT_STD_NONLINEAR,
}[flags.is_linear]


def _rekey_dict(dict_to_rekey: dict, old_key_to_new_key: dict, /, garbage_key=None) -> dict:
    #[
    new_dict = {
        old_key_to_new_key.get(key, garbage_key): value
        for key, value in dict_to_rekey.items()
    }
    if garbage_key in new_dict:
        del new_dict[garbage_key]
    return new_dict
    #]


def resolve_variant(self, variants, /, ) -> Iterable[int]:
    #[
    if isinstance(variants, Number):
        return [variants, ]
    elif variants is Ellipsis:
        return range(self.num_variants)
    elif isinstance(variants, slice):
        return range(*variants.indices(self.num_variants))
    else:
        return [v for v in variants]
    #]





"""
Jacobians for steady systems
"""


#[
from __future__ import annotations

from typing import (Protocol, )
import numpy as _np
import scipy as _sp

from ..incidences import main as _incidence

from . import _base
#]



class SteadyJacobianProtocol(Protocol, ):
    """
    """
    #[
    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        ...
    #]


class _AtomFactory:
    """
    """
    #[
    def create_diff_for_token(*args, **kwargs):
        """
        """
        ...

    @staticmethod
    def create_data_index_for_token(
        token: _incidence.Token,
        columns_to_eval: tuple[int, int],
        /,
    ) -> tuple[int, slice]:
        """
        """
        return (
            token.qid,
            slice(columns_to_eval[0], columns_to_eval[1]+1),
        )
    #]


class _FlatAtomFactory(_AtomFactory, ):
    """
    """
    #[
    @staticmethod
    def create_diff_for_token(
        token: _incidence.Token,
        wrt_qids: tuple[int],
        /,
    ) -> _np.ndarray:
        """
        """
        try:
            index = wrt_qids.index(token.qid)
            diff = _np.zeros((len(wrt_qids), 1))
            diff[index] = 1
            return diff
        except:
            return 0
    #]


class _NonflatAtomFactory(_AtomFactory, ):
    """
    """
    #[
    @staticmethod
    def create_diff_for_token(
        token: _incidence.Token,
        wrt_qids: dict[int, tuple[int]],
        /,
    ) -> _np.ndarray:
        """
        """
        try:
            index = wrt_qids.index(token.qid)
            diff = _np.zeros((len(wrt_qids), 2))
            diff[index, :] = 1, token.shift
            return diff
        except:
            return 0
    #]


class FlatSteadyJacobian(_base.Jacobian, ):
    """
    """
    #[
    atom_factory = _FlatAtomFactory

    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        diff_array = self._aldi_context.eval_diff_to_array(steady_array, steady_array, )
        return self._create_jacobian(self._shape, diff_array, self._map, )
    #]


class NonflatSteadyJacobian(_base.Jacobian, ):
    """
    """
    #[
    NONFLAT_STEADY_SHIFT: int = 1
    atom_factory = _NonflatAtomFactory

    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        diff_array = self._aldi_context.eval_diff_to_array(steady_array, steady_array, )
        A = self._create_jacobian(self._shape, diff_array[:, 0:1], self._map, )
        B = self._create_jacobian(self._shape, diff_array[:, 1:2], self._map, )
        k = self.NONFLAT_STEADY_SHIFT
        return _np.vstack((
            _np.hstack((A, B, )),
            _np.hstack((A, B+k*A, )),
        ))
    #]





"""
Steady equator
"""


#[
from __future__ import annotations

import numpy as _np
from typing import (Callable, Protocol, )
from collections.abc import (Iterable, )

from .. import equations as _equations
from . import plain as _plain
#]


class SteadyEquatorProtocol:
    """
    """
    #[
    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        ...
    #]


class SteadyEquator:
    """
    """
    #[
    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        t_zero: int,
        /,
        *,
        custom_functions: dict[str, Callable] | None = None,
    ) -> None:
        self._equator = _plain.PlainEquator(equations, custom_functions=custom_functions, )
        self._t_zero = t_zero

    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        ...
    #]


class FlatSteadyEquator(SteadyEquator, ):
    """
    """
    #[
    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        return self._equator.eval(steady_array, self._t_zero, steady_array, )
    #]


class NonflatSteadyEquator(SteadyEquator, ):
    """
    """
    #[
    NONFLAT_STEADY_SHIFT: int = 1

    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        k = self.NONFLAT_STEADY_SHIFT
        return _np.hstack((
            self._equator.eval(steady_array, self._t_zero, steady_array, ),
            self._equator.eval(steady_array, self._t_zero+k, steady_array, ),
        ))
    #]





"""
"""


#[
from __future__ import annotations

from typing import (Self, Callable, )

from .. import equations as _equations
from .. import quantities as _quantities
from .. import wrongdoings as _wrongdoings
from ..fords import descriptors as _descriptors
from ..equators import plain as _equators

from . import _flags
#]


_PLAIN_EQUATOR_EQUATION = (
    _equations.EquationKind.TRANSITION_EQUATION
    | _equations.EquationKind.MEASUREMENT_EQUATION
)


class Invariant:
    """
    Invariant part of a Model object
    """
    #[
    __slots__ = (
        "_flags",
        "_function_context",
        "_quantities",
        "_dynamic_equations",
        "_steady_equations",
        "_dynamic_descriptor",
        "_steady_descriptor",
        "_min_shift",
        "_max_shift",
    )
    def __init__(
        self,
        model_source,
        /,
        context: dict | None = None,
        check_syntax: bool = True,
        **kwargs,
    ) -> Self:
        """
        """
        self._flags = _flags.Flags.from_kwargs(**kwargs, )
        #
        self._populate_function_context(context)
        #
        self._quantities = _cp.deepcopy(model_source.quantities, )
        self._dynamic_equations = _cp_deepcopy(model_source.dynamic_equations, )
        self._steady_equations = _cp.deepcopy(model_source.steady_equations, )
        #
        name_to_qid = _quantities.create_name_to_qid(self._quantities, )
        _equations.finalize_dynamic_equations(self._dynamic_equations, name_to_qid, )
        _equations.finalize_steady_equations(self._steady_equations, name_to_qid, )
        #
        if check_syntax:
            _check_syntax(self._dynamic_equations, self._function_context, )
            _check_syntax(self._steady_equations, self._function_context, )
        #
        self._dynamic_descriptor = _descriptors.Descriptor(self._dynamic_equations, self._quantities, self._function_context, )
        self._steady_descriptor = _descriptors.Descriptor(self._steady_equations, self._quantities, self._function_context, )
        #
        self._plain_equator_for_dynamic_equations = _equators.PlainEquator(
            _equations.generate_equations_of_kind( self._dynamic_equations, _PLAIN_EQUATOR_EQUATION, ),
            custom_functions=self._function_context,
        )
        #
        self._plain_equator_for_steady_equations = _equators.PlainEquator(
            _equations.generate_equations_of_kind(self._steady_equations, _PLAIN_EQUATOR_EQUATION, ),
            custom_functions=self._function_context,
        )
        #
        self._min_shift, self._max_shift = None, None
        self._populate_min_max_shifts()

    def _populate_min_max_shifts(self, /, ) -> None:
        """
        """
        self._min_shift = _equations.get_min_shift_from_equations(
            self._dynamic_equations + self._steady_equations,
        )
        self._max_shift = _equations.get_max_shift_from_equations(
            self._dynamic_equations + self._steady_equations,
        )

    def _populate_function_context(
        self,
        context: dict | None,
        /,
    ) -> None:
        """
        """
        self._function_context = {
            k: v for k, v in context.items()
            if isinstance(v, Callable)
        } if context else None
    #]


def _check_syntax(equations, function_context, /, ):
    """
    Try all equations at once; if this fails, do equation by equation to # catch the troublemakers
    """
    try:
        eval(_equations.create_equator_func_string(equations), )
    except:
        _catch_troublemakers(equations, function_context, )
    #]


def _catch_troublemakers(equations, function_context, /, ):
    """
    Catch the troublemakers
    """
    #[
    fail = [
        eqn.human for eqn in equations
        if not _success_creating_lambda(eqn, function_context)
    ]
    if fail:
        message = ["Syntax error in these equations"] + fail
        _wrongdoings.throw("error", message)
    #]


def _success_creating_lambda(equation, function_context):
    """
    """
    #[
    try:
        eval(_equations.create_equator_func_string([equation.xtring]))
        return True
    except Exception as ex:
        return False
    #]




"""
Algorithmic differentiator
"""


#[
from __future__ import annotations

from typing import (Self, Callable, Protocol, )
from numbers import (Number, )
from collections.abc import (Iterable, Sequence, )
import numpy as _np
import copy as _cp

from ..exceptions import ListException

from ..aldi import finite_differentiators as af_
from ..aldi import adaptations as aa_
from ..incidences import main as _incidence
from .. import equations as _equations
#]


class ValueMixin:
    #[
    @property
    def value(self):
        return self._value if self._value is not None else type(self)._data_context[self._data_index]
    #]


class Atom(ValueMixin, ):
    """
    Atomic value for differentiation
    """
    _data_context: _np.ndarray | None = None
    _is_atom: bool = True
    #[
    def __init__(self) -> None:
        """
        """
        self._value = None
        self._diff = None
        self._logly = None
        self._data_index = None

    @classmethod
    def no_context(
        cls,
        value: Number,
        diff: Number,
        /,
        logly: bool | None = False,
    ) -> Self:
        """
        Create atom with self-contained data and logly contexts
        """
        self = cls()
        self._value = value
        self._diff = diff
        self._logly = logly if logly is not None else False
        return self

    @classmethod
    def in_context(
        cls,
        /,
        diff: _np.ndarray | Number,
        data_index: tuple[int, slice],
        logly: bool,
    ) -> Self:
        """
        Create atom with pointers to data and logly contexts
        """
        self = Atom()
        self._diff = diff
        self._data_index = data_index
        self._logly = logly if logly is not None else False
        return self

    @classmethod
    def zero(
        cls,
        diff_shape: tuple[int, int],
    ) -> Self:
        return cls.no_context(0, _np.zeros(diff_shape), False)

    @property
    def diff(self):
        return self._diff if not self._logly else self._diff * self.value

    def __pos__(self):
        return self

    def __neg__(self):
        new_value = -self.value
        new_diff = -self.diff
        return type(self).no_context(new_value, new_diff, False)

    def __add__(self, other):
        if hasattr(other, "_is_atom"):
            new_value = self.value + other.value
            new_diff = self.diff + other.diff
        else:
            new_value = self.value + other
            new_diff = self.diff
        return type(self).no_context(new_value, new_diff, False)

    def __sub__(self, other):
        if hasattr(other, "_is_atom"):
            new_value = self.value - other.value
            new_diff = self.diff - other.diff
        else:
            new_value = self.value - other
            new_diff = self.diff
        new_logly = False
        return type(self).no_context(new_value, new_diff, False)

    def __mul__(self, other):
        self_value = self.value
        self_diff = self.diff
        if hasattr(other, "_is_atom"):
            other_value = other.value
            other_diff = other.diff
            new_value = self_value * other_value
            new_diff = self_diff * other_value + self_value * other_diff
        else:
            new_value = self_value * other
            new_diff = self_diff * other
        return type(self).no_context(new_value, new_diff, False)

    def __truediv__(self, other):
        self_value = self.value
        self_diff = self.diff
        if hasattr(other, "_is_atom"):
            other_value = other.value
            other_diff = other.diff
            new_value = self_value / other_value
            new_diff = (self_diff*other_value - self_value*other_diff) / (other_value**2)
        else:
            new_value = self_value / other
            new_diff = self_diff / other
        return type(self).no_context(new_value, new_diff, False)

    def __rtruediv__(self, other):
        self_value = self.value
        self_diff = self.diff
        new_value = other / self_value
        new_diff = -other*self_diff / (self_value**2)
        return type(self).no_context(new_value, new_diff, False)

    def __pow__(self, other):
        """
        Differenatiate self**other 
        """
        if hasattr(other, "_is_atom"):
            # self(x)**other(x)
            # d[self(x)**other(x)] = d[self(x)**other] + d[self**other(x)]
            new_value = self.value**other.value
            _, new_diff_exp = self._power(other.value) # self(x)**other
            _, new_diff_pow = other._exponential(self.value) # self**other(x)
            new_diff = new_diff_exp + new_diff_pow
        else:
            # self(x)**other
            new_value, new_diff = self._power(other)
        return type(self).no_context(new_value, new_diff, False)

    def _exponential(self, other_value):
        """
        Differenatiate exponential function other_value**self(x)
        """
        new_value = other_value**self.value
        new_diff = (
            other_value**self.value * _np.log(other_value) * self.diff 
            if other_value != 0 else 0
        )
        return new_value, new_diff

    def _power(self, other_value):
        """
        Differenatiate power function self(x)**other_value
        """
        self_value = self.value
        self_diff = self.diff
        new_value = self_value ** other_value
        new_diff = other_value * (self_value**(other_value-1)) * self_diff
        return new_value, new_diff

    __rmul__ = __mul__

    __radd__ = __add__

    def __rsub__(self, other, /, ) -> Self:
        """
        Differenatiate other - self
        """
        return self.__neg__().__add__(other)

    def _log_(self, /, ) -> Self:
        """
        Differentiate log(self)
        """
        new_value = _np.log(self.value)
        new_diff = 1 / self.value * self.diff
        return type(self).no_context(new_value, new_diff, False)

    def _exp_(self, /, ) -> Self:
        """
        Differentiate exp(self)
        """
        new_value = _np.exp(self.value)
        new_diff = new_value * self.diff
        return type(self).no_context(new_value, new_diff, False)

    def _sqrt_(self, /, ) -> Self:
        """
        Differentiate sqrt(self)
        """
        new_value = _np.sqrt(self.value)
        new_diff = 0.5 / _np.sqrt(self.diff)
        return type(self).no_context(new_value, new_diff, False)

    def _maximum_(
        self,
        floor: Number = 0,
        /, 
    ) -> Self:
        """
        Differenatiate maximum(self, floor)
        """
        orig_value = self.value
        orig_diff = self.diff
        if isinstance(orig_value, Number) and isinstance(orig_diff, Number):
            orig_value = _np.array(orig_value, dtype=float)
            orig_diff = _np.array(orig_diff, dtype=float)
        new_value = _np.copy(orig_value)
        inx_floor = orig_value == floor
        inx_below = orig_value < floor
        inx_above = orig_value > floor
        new_value[inx_below] = floor
        new_diff = _np.copy(orig_diff)
        multiplier = _np.copy(orig_value)
        multiplier[inx_floor] = 0.5
        multiplier[inx_above] = 1
        multiplier[inx_below] = 0
        new_diff = orig_diff * multiplier
        return type(self).no_context(new_value, new_diff, False)

    def _mininum_(
        self,
        ceiling: Number = 0,
        /,
    ) -> Self:
        """
        Differenatiate minimum(self, ceiling)
        """
        return (-self)._max_(-ceiling)
    #]


class Context:
    """
    """
    #[
    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        atom_factory: AtomFactoryProtocol,
        /,
        eid_to_wrts: dict[int, tuple[Any]],
        qid_to_logly: dict[int, bool] | None,
        num_columns_to_eval: int,
        custom_functions: dict | None,
    ) -> Self:
        """
        """
        self._eid_to_wrts = eid_to_wrts
        self._qid_to_logly = qid_to_logly or {}
        self._num_columns_to_eval = num_columns_to_eval
        self._populate_equations(equations, )
        #
        all_tokens = set(_equations.generate_all_tokens_from_equations(self._equations, ), )
        self.min_shift = _incidence.get_min_shift(all_tokens, )
        self.max_shift = _incidence.get_max_shift(all_tokens, )
        self.shape_data = (
            1 + (_incidence.get_max_qid(all_tokens) or 0),
            -self.min_shift + num_columns_to_eval + self.max_shift,
        )
        #
        self._populate_atom_dict(atom_factory, )
        #
        custom_functions = { 
            k: af_.finite_differentiator(v) 
            for k, v in custom_functions.items()
        } if custom_functions else None
        custom_functions = aa_.add_function_adaptations_to_custom_functions(custom_functions)
        custom_functions["Atom"] = Atom
        #
        self._equator = _ep.PlainEquator(self._equations, custom_functions, )

    @property
    def _t_zero(self, /, ) -> int:
        """
        """
        return -self.min_shift

    def _get_num_wrts(self, eid, /, ) -> int:
        """
        """
        return len(self._eid_to_wrts[eid])

    def _get_columns_to_eval(self, /, ) -> tuple[int, int]:
        """
        """
        return self._t_zero, self._t_zero + self._num_columns_to_eval - 1

    def _get_diff_shape_for_eid(self, eid, /, ) -> tuple[int, int]:
        """
        """
        return self._get_num_wrts(eid), self._num_columns_to_eval

    def _populate_equations(
        self,
        equations: Iterable[_equations.Equation],
        /,
    ) -> None:
        """
        """
        self._equations = tuple(
            _adapt_equation_for_aldi(e, _self._get_diff_shape_for_eid(e.id), )
            for e in equations
        )

    def _populate_atom_dict(
        self,
        atom_factory: AtomFactoryProtocol,
        /,
    ) -> None:
        """
        * atom_factory -- Implement create_diff_for_token, create_data_index_for_token, get_logly_for_token
        * qid_to_logly -- Dictionary of qid to logly
        * columns_to_eval -- Tuple of (first, last) column indices to be evaluated
        """
        columns_to_eval = self._get_columns_to_eval()
        self._x = {}
        for eqn in self._equations:
            for tok in eqn.incidence:
                atom = Atom.in_context(
                    diff=atom_factory.create_diff_for_token(tok, self._eid_to_wrts[eqn.id], ),
                    data_index=atom_factory.create_data_index_for_token(tok, columns_to_eval, ),
                    logly=self._qid_to_logly.get(tok.qid, False, ),
                )
                self._x[eqn.id][(tok.qid, tok.shift)] = atom

    def eval(
        self,
        data_array: _np.ndarray,
        steady_array: _np.ndarray,
    ) -> Iterable[Atom]:
        """
        Evaluate and return the list of final atoms, one atom for each equation
        """
        self._verify_data_array_shape(data_array.shape, )
        Atom._data_context = data_array
        output = self._equator.eval(self._x, None, steady_array, )
        Atom._data_context = None
        return output

    def eval_to_arrays(
        self,
        *args,
    ) -> tuple[_np.ndarray, _np.ndarray]:
        """
        Evaluate and return arrays of diffs and values extracted from final atoms
        """
        output = self.eval(*args)
        diff = _np.vstack([x.diff for x in output])
        value = _np.vstack([x.value for x in output])
        return diff, value

    def eval_diff_to_array(
        self,
        *args,
    ) -> _np.array:
        """
        Evaluate and return array of diffs
        """
        return _np.vstack([
            x.diff for x in self.eval(*args)
            if hasattr(x, "diff")
        ])

    def _verify_data_array_shape(self, shape_data: _np.ndarray) -> None:
        """
        """
        if shape_data[0]>=self.shape_data[0] and shape_data[1]>=self.shape_data[1]:
            return
        raise InvalidInputDataArrayShape(shape_data, self.shape_data)
    #]


def _adapt_equation_for_aldi(
    equation: _eq.Equation,
    diff_shape: tuple[int, int],
    /,
) -> _equations.Equation:
    """
    """
    #[
    aldi_equation = _cp.deepcopy(equation)
    aldi_equation.xtring = aldi_equation.xtring.replace("x[", f"x[{eid}][")
    aldi_equation.xtring += f" + Atom.zero({diff_shape}, )"
    return aldi_equation
    #]


class InvalidInputDataArrayShape(ListException, ):
    """
    """
    #[
    def __init__(
        self,
        shape_entered: tuple[int, int],
        needed: tuple[int, int]
    ) -> None:
        messages = [ f"Incorrect size of input data matrix: entered {shape_entered}, needed {needed}" ]
        super().__init__(messages)
    #]


class AtomFactoryProtocol(Protocol, ):
    """
    Protocol for creating atoms representing tokens
    """
    #[
    def create_diff_for_token(
        self,
        token: _incidence.Token,
        wrts: tuple[Any, ...],
    ) -> _np.ndarray:
        """
        Create a diff for an atom representing a given token
        """
        ...

    def create_data_index_for_token(
        self,
        token: _incidence.Token,
        columns_to_eval: tuple[int, int],
    ) -> tuple[int, slice]:
        """
        Create a data index for an atom representing a given token
        """
        ...
    #]





"""
"""


#[
from __future__ import (annotations, )

import enum as en_
#]


class Flags(en_.IntFlag, ):
    """
    """
    #[
    DEFAULT = 0
    LINEAR = en_.auto()
    FLAT = en_.auto()

    @property
    def is_linear(self, /, ) -> bool:
        return Flags.LINEAR in self

    @property
    def is_nonlinear(self, /, ) -> bool:
        return not self.is_linear

    @property
    def is_flat(self, /, ) -> bool:
        return Flags.FLAT in self

    @property
    def is_nonflat(self, /, ) -> bool:
        return not self.is_flat

    def update_from_kwargs(self, /, **kwargs) -> Self:
        linear = kwargs.get("linear") if kwargs.get("linear") is not None else self.is_linear
        flat = kwargs.get("flat") if kwargs.get("flat") is not None else self.is_flat
        return type(self).from_kwargs(linear=linear, flat=flat)

    @classmethod
    def from_kwargs(cls: type, **kwargs, ) -> Self:
        self = cls.DEFAULT
        if kwargs.get("linear"):
            self |= cls.LINEAR
        if kwargs.get("flat"):
            self |= cls.FLAT
        return self
    #]





"""
Plotly interface to time series objects
"""


#[
from __future__ import annotations

from types import EllipsisType
import plotly.graph_objects as _pg
import plotly.subplots as _ps

from .. import dates as _dates
#]


builtin_range = range


_COLOR_ORDER = [
  "rgb(  0,   114,   189)",
  "rgb(217,    83,    25)",
  "rgb(237,   177,    32)",
  "rgb(126,    47,   142)",
  "rgb(119,   172,    48)",
  "rgb( 77,   190,   238)",
  "rgb(162,    20,    47)",
]


class Mixin:
    """
    """
    #[
    def plot(
        self,
        *,
        range: Iterable[_dates.Dater] | EllipsisType = ...,
        title: str | None = None,
        legend: Iterable[str] | None = None,
        show_legend: bool | None = None,
        figure = None,
        subplot: tuple[int, int] | int | None = None,
        xline = None,
    ) -> _pg.Figure:
        range = self._resolve_dates(range)
        range = [ t for t in range ]
        data = self.get_data(range)
        num_columns = data.shape[1]
        date_str = [ t.to_plotly_date() for t in range ]
        date_format = range[0].frequency.plotly_format
        figure = _pg.Figure() if figure is None else figure
        subplot = _resolve_subplot(figure, subplot, )
        for i in builtin_range(num_columns):
            figure.add_trace(_pg.Scatter(
                x=date_str, y=data[:, i], name=legend[i] if legend else None,
                line={"color": _COLOR_ORDER[i % len(_COLOR_ORDER)]},
            ), row=subplot[0], col=subplot[1])
        show_legend = show_legend if show_legend is not None else legend is not None
        figure.update_layout(title=title, xaxis={"tickformat": date_format}, showlegend=show_legend, )
        if xline:
            xline = xline.to_plotly_date()
            figure.add_vline(xline)
        return figure
    #]


def _resolve_subplot(
    figure: _pg.Figure,
    subplot: tuple[int, int] | int | None,
    /,
) -> tuple[int, int]:
    """
    """
    if subplot is None:
        return None, None
    if isinstance(subplot, tuple) or isinstance(subplot, list):
        return subplot[0], subplot[1]
    if isinstance(subplot, int):
        position = subplot
        rows, columns = figure._get_subplot_rows_columns()
        num_rows = len(rows)
        num_columns = len(columns)
        row = position // num_columns + 1
        column = position % num_columns + 1
        return row, column




"""
Simulation plans
"""


#[
from __future__ import annotations

from typing import (Iterable, Protocol, )

from .. import dates as _dates
from .. import wrongdoings as _wrongdoings
from .. import transforms as _transforms
#]


__all__ = (
    "Plan",
)


class PlannableProtocol(Protocol, ):
    """
    """
    #[
    def can_be_exogenized(self, /, ) -> Iterable[str]: ...
    def can_be_endogenized(self, /, ) -> Iterable[str]: ...
    #]


class Plan:
    """
    """
    __slots__ = (
        "base_range",
        "anticipate",
        "exogenized",
        "endogenized",
    )

    def __init__(
        self,
        plannable,
        base_range: Iterable[Dater],
        /,
        anticipate: bool = True,
    ) -> None:
        """
        """
        self.base_range = tuple(base_range)
        self.anticipate = anticipate
        self._initialize(plannable, )

    @property
    def start_date(self, /, ) -> Dater:
        """
        """
        return self.base_range[0]

    @property
    def num_periods(self, /, ) -> int:
        """
        """
        return len(self.base_range)

    @property
    def can_be_exogenized(self, /, ) -> tuple[str]:
        """
        """
        return tuple(self.exogenized.keys())

    @property
    def can_be_endogenized(self, /, ) -> tuple[str]:
        """
        """
        return tuple(self.endogenized.keys())

    def _initialize(
        self,
        plannable: PlannableProtocol,
        /,
    ) -> None:
        """
        """
        can_be_exogenized = plannable.can_be_exogenized()
        self.exogenized = {
            n: [None] * self.num_periods
            for n in can_be_exogenized
        } if can_be_exogenized else {}
        #
        can_be_endogenized = plannable.can_be_endogenized()
        self.endogenized = {
            n: [None] * self.num_periods
            for n in can_be_endogenized
        } if can_be_endogenized else {}

    def exogenize(
        self,
        dates: Iterable[_dates.Dater] | Ellipsis,
        names: Iterable[str] | str,
        /,
        transform: str | None = None,
        when_data: bool = False,
    ) -> None:
        """
        """
        names = self._resolve_and_check_names(self.exogenized, names, "exogenized", )
        date_indexes = self._resolve_dates(dates, )
        transform = _transforms.RESOLVE_TRANSFORM[transform](when_data, )
        for n in names:
            for t in date_indexes:
                self.exogenized[n][t] = transform

    def collect_databank_names(self, /, ) -> tuple[str]:
        """
        """
        databank_names = set()
        for k, v in self.exogenized.items():
            databank_names.update(
                t.resolve_databank_name(k, )
                for t in v if t is not None
            )
        return tuple(databank_names)

    def _resolve_and_check_names(
        self,
        what: dict,
        names: Iterable[str] | str,
        done: str,
        /,
    ) -> tuple[str]:
        """
        """
        if names is Ellipsis:
            return tuple(what.keys())
        if isinstance(names, str):
            names = (names, )
        names = tuple(names)
        invalid = [n for n in names if n not in what]
        if invalid:
            raise _wrongdoings.IrisPieError(
                [f"These name(s) cannot be {done}:"] + invalid
            )
        return names

    def _resolve_dates(
        self,
        dates,
        /,
    ) -> tuple[int, ...]:
        """
        """
        if dates is Ellipsis:
            return tuple(range(self.num_periods))
        invalid = [repr(d) for d in dates if d not in self.base_range]
        if invalid:
            raise _wrongdoings.IrisPieError(
                ["These date(s) are out of simulation range:"] + invalid
            )
        return tuple(d - self.start_date for d in dates)




"""
Calculate first derivatives of a custom function using finite approximation
"""


#[
from __future__ import annotations

from typing import (Callable, )
import numpy as np_
import copy as co_

from ..aldi import (differentiators as ad_, )
#]


#••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••
# Exposure
#••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••


def finite_differentiator(func: Callable) -> Callable:
    """
    Decorate a custom function for finite differentiation
    """
    def wrapper(*args):
        return _calculate_finite_derivatives(func, *args)
    return wrapper


#••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••
# Implementation
#••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••


_RELATIVE_FINITE_DIFF_STEP = 1e-6


def _calculate_finite_derivatives(
    func: Callable,
    /,
    *args,
) -> Callable:
    """
    Total finite differentiation of a custom function
    """
    arg_values = _collect_arg_values(*args, )
    new_value = func(*arg_values, )
    arg_diffs = _collect_arg_diffs(*args, )
    new_diff = sum(
        _partial_times_inner(func, k, arg_values, arg_diffs, )
        for k in range(len(args, ), )
    )
    return ad_.Atom.no_context(new_value, new_diff, False, )


def _partial_times_inner(func, k, arg_values, arg_diffs, ):
    """
    For f(..., g(x), ...), evaluate diff of f wrt the k-th argument times dg/dx
    """
    return (
        _partial_two_sided_derivative(func, k, arg_values, ) * arg_diffs[k]
        if arg_diffs[k] is not None else 0
    )


def _partial_two_sided_derivative(func, k, arg_values, ):
    """
    For f(..., x, ...), evaluate diff of f wrt the k-th argument
    """
    epsilon = _get_epsilon(arg_values[k])
    arg_values_plus = _plus_epsilon(arg_values, k, epsilon, )
    arg_values_minus = _plus_epsilon(arg_values, k, -epsilon, )
    return (func(*arg_values_plus) - func(*arg_values_minus)) / (2 * epsilon)


def _plus_epsilon(arg_values, k, epsilon, ):
    """
    Create a copy of function arguments and increase k-th argument by epsilon
    """
    arg_values_plus = co_.deepcopy(arg_values)
    arg_values_plus[k] += epsilon
    return arg_values_plus


def _get_epsilon(value, ):
    """
    Calculate the differentiation step based on the value around which we differentiate
    """
    base = np_.maximum(abs(value), 1)
    return base * _RELATIVE_FINITE_DIFF_STEP


def _collect_arg_values(*args, ):
    """
    Collect the values of input arguments, both Atom values and primitives
    """
    return [
        co_.deepcopy(a.value) if hasattr(a, "_is_atom", ) else a
        for a in args
    ]


def _collect_arg_diffs(*args, ):
    """
    Collect the diffs of input arguments for Atoms or Nones for primitives
    """
    return [
        co_.deepcopy(a.diff) if hasattr(a, "_is_atom", ) else None
        for a in args
    ]





"""
First-order system simulators
"""


#[
from __future__ import annotations

from typing import (Self, TypeAlias, Literal, Protocol, runtime_checkable)
from collections.abc import (Iterable, )
import numpy as _np

from ..databanks import main as _databanks
from ..fords import simulators as _simulators
from .. import dataslabs as _dataslabs
#]


class SimulateMixin:
    """
    """
    #[
    def simulate(
        self,
        in_databank: _databanks.Databank,
        base_range: Iterable[Dater],
        /,
        anticipate: bool = True,
        deviation: bool = False,
        prepend_input: bool = True,
    ) -> _databanks.Databank:
        """
        """
        dataslabs = tuple(
            _dataslabs.Dataslab.from_databank_for_simulation(self, in_databank, base_range, column=i, )
            for i in range(self.num_variants)
        )

        for variant, dataslab in zip(self._variants, dataslabs):
            new_data = dataslab.copy_data()
            new_data = _simulators.simulate_flat(
                variant.solution, self.get_solution_vectors(),
                new_data, dataslab.base_columns, deviation, anticipate,
            )
            dataslab.data = new_data
            dataslab.remove_columns(dataslab.base_columns[-1] - dataslab.num_ext_periods + 1)

        out_databank = _dataslabs.multiple_to_databank(dataslabs)
        if prepend_input:
            out_databank.prepend(in_databank, base_range[0]-1, )

        out_databank = out_databank | self.get_parameters_stds()

        return out_databank
    #]





"""
# Descriptor for first-order systems and solutions

## Unsolved system

$$
A E[x_{t}] + B E[x_{t-1}] + C + D v_{t} = 0 \\
F y_{t} + G x_{t} + H + J w_{t} = 0
$$

## State-space solution

$$
x_{t} = T x_{t-1} + K + R v_{t} \\
y_{t} = Z x_{t} + D + H w_{t}
$$
"""


#[
from __future__ import annotations

from typing import (Self, )
from collections.abc import (Iterable, )
import itertools as it_
import dataclasses as _dc
import numpy as np_

from ..incidences import main as _incidence
from .. import equations as _equations
from .. import quantities as _quantities
from ..aldi import differentiators as _differentiators
from ..aldi import maps as _maps
#]


class _AtomFactory:
    """
    """
    #[
    @staticmethod
    def create_diff_for_token(
        token: Token,
        wrt_tokens: tuple[Token, ...],
        /,
    ) -> np_.ndarray:
        """
        """
        try:
            index = wrt_tokens.index(token)
            diff = np_.zeros((len(wrt_tokens), 1))
            diff[index] = 1
            return diff
        except:
            return 0

    @staticmethod
    def create_data_index_for_token(
        token: Token,
        columns_to_eval: tuple[int, slice],
    ) -> tuple[int, slice]:
        """
        """
        return (
            token.qid,
            slice(columns_to_eval[0]+token.shift, columns_to_eval[1]+token.shift+1),
        )
    #]


@_dc.dataclass(slots=True, )
class Descriptor:
    """
    """
    #[
    system_vectors: SystemVectors | None = None
    solution_vectors: SolutionVectors | None = None
    system_map: SystemMap | None = None
    aldi_context: _differentiators.Context | None = None

    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        quantities: Iterable[_quantities.Quantity],
        custom_functions: dict | None,
        /,
    ) -> None:
        self.system_vectors = SystemVectors(equations, quantities)
        self.solution_vectors = SolutionVectors(self.system_vectors)
        self.system_map = SystemMap(self.system_vectors)
        #
        # Create the context for the algorithmic differentiator
        system_equations = _select_system_equations_from_equations(
            equations,
            self.system_vectors.transition_eids,
            self.system_vectors.measurement_eids,
        )
        #
        # Create the evaluation context for the algorithmic differentiator
        self.aldi_context = _differentiators.Context(
            system_equations,
            _AtomFactory,
            eid_to_wrts=self.system_vectors.eid_to_wrt_tokens,
            qid_to_logly=_quantities.create_qid_to_logly(quantities),
            num_columns_to_eval=1,
            custom_functions=custom_functions,
        )

    def get_num_backwards(self: Self) -> int:
        return self.system_vectors.get_num_backwards()

    def get_num_forwards(self: Self) -> int:
        return self.system_vectors.get_num_forwards()
    #]


"""
Quantity kinds that can be used in the transition part of a first-order system
"""
_TRANSITION_SYSTEM_QUANTITY = (
    _quantities.QuantityKind.TRANSITION_VARIABLE
    | _quantities.QuantityKind.TRANSITION_SHOCK
)


"""
Quantity kinds that can be used in the measurement part of a first-order system
"""
_MEASUREMENT_SYSTEM_QUANTITY = (
    _quantities.QuantityKind.MEASUREMENT_VARIABLE
    | _quantities.QuantityKind.TRANSITION_VARIABLE
    | _quantities.QuantityKind.MEASUREMENT_SHOCK
)


"""
Quantity kinds that can be used in a first-order system
"""
_SYSTEM_QUANTITY = (
    _TRANSITION_SYSTEM_QUANTITY
    | _MEASUREMENT_SYSTEM_QUANTITY
)


@_dc.dataclass(slots=True, )
class SystemVectors:
    """
    Vectors of quantities and equation ids in first-order system matrices
    """
    #[
    transition_eids: Iterable[int] | None = None
    measurement_eids: Iterable[int] | None = None
    eid_to_wrt_tokens: dict[int, Iterable[_incidence.Token]] | None = None
    #
    transition_variables: Iterable[_incidence.Token] | None = None
    initial_conditions: Iterable[bool] | None = None,
    transition_shocks: Iterable[_incidence.Token] | None = None 
    measurement_variables: Iterable[_incidence.Token] | None = None
    measurement_shocks: Iterable[_incidence.Token] | None = None 
    #
    shape_A_excl_dynid: tuple[int, int] | None = None
    shape_B_excl_dynid: tuple[int, int] | None = None
    shape_C_excl_dynid: tuple[int, int] | None = None
    shape_D_excl_dynid: tuple[int, int] | None = None
    shape_F: tuple[int, int] | None = None
    shape_G: tuple[int, int] | None = None
    shape_H: tuple[int, int] | None = None
    shape_J: tuple[int, int] | None = None

    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        quantities: Iterable[_quantities.Quantity],
        /,
    ) -> None:
        """
        Construct system vectors from a list of equations and a list of quantities
        """
        self.transition_eids = sorted([eqn.id for eqn in equations if eqn.kind in _equations.EquationKind.TRANSITION_EQUATION])
        self.measurement_eids = sorted([eqn.id for eqn in equations if eqn.kind in _equations.EquationKind.MEASUREMENT_EQUATION])
        qid_to_kind = _quantities.create_qid_to_kind(quantities)
        all_tokens = set(_equations.generate_all_tokens_from_equations(equations))
        all_wrt_tokens = set(_incidence.generate_tokens_of_kinds(all_tokens, qid_to_kind, _SYSTEM_QUANTITY))
        self.eid_to_wrt_tokens = _equations.create_eid_to_wrt_tokens(equations, all_wrt_tokens)
        #
        actual_tokens_transition_variables = set(_incidence.generate_tokens_of_kinds(all_tokens, qid_to_kind, _quantities.QuantityKind.TRANSITION_VARIABLE))
        #
        # Make adjustment for transition variables in measurement
        # equations: each x(t-k) in measurement needs to be in the current
        # dated (LHS) vector of transition variables; this is done by
        # pretending x(t-k-1) is needed
        adjusted_tokens_transition_variables = _adjust_for_measurement_equations(actual_tokens_transition_variables, equations, qid_to_kind)
        #
        self.transition_variables = _incidence.sort_tokens(_create_system_transition_vector(adjusted_tokens_transition_variables))
        self.initial_conditions = [
            _incidence.Token(t.qid, t.shift-1) in actual_tokens_transition_variables and t.shift <= 0
            for t in self.transition_variables
        ]
        self.transition_shocks = _incidence.sort_tokens(
            _incidence.generate_tokens_of_kinds(all_tokens, qid_to_kind, _quantities.QuantityKind.TRANSITION_SHOCK)
        )
        #
        self.measurement_variables = _incidence.sort_tokens(
            _incidence.generate_tokens_of_kinds(all_tokens, qid_to_kind, _quantities.QuantityKind.MEASUREMENT_VARIABLE)
        )
        self.measurement_shocks = _incidence.sort_tokens(
            _incidence.generate_tokens_of_kinds(all_tokens, qid_to_kind, _quantities.QuantityKind.MEASUREMENT_SHOCK)
        )
        #
        self.shape_A_excl_dynid = (len(self.transition_eids), len(self.transition_variables))
        self.shape_B_excl_dynid = self.shape_A_excl_dynid
        self.shape_C_excl_dynid = (len(self.transition_eids), 1)
        self.shape_D_excl_dynid = (len(self.transition_eids), len(self.transition_shocks))
        #
        self.shape_F = (len(self.measurement_eids), len(self.measurement_variables))
        self.shape_G = (len(self.measurement_eids), len(self.transition_variables))
        self.shape_H = (len(self.measurement_eids), 1)
        self.shape_J = (len(self.measurement_eids), len(self.measurement_shocks))

    def get_num_backwards(self) -> int:
        return _get_num_backwards(self.transition_variables)

    def get_num_forwards(self) -> int:
        return _get_num_forwards(self.transition_variables)
    #]


@_dc.dataclass(slots=True, )
class SolutionVectors:
    """
    Vectors of quantities in first-order solution matrices
    """
    #[
    transition_variables: Iterable[_incidence.Token] | None = None
    initial_conditions: Iterable[bool] | None = None,
    transition_shocks: Iterable[_incidence.Token] | None = None 
    measurement_variables: Iterable[_incidence.Token] | None = None
    measurement_shocks: Iterable[_incidence.Token] | None = None 

    def __init__(self, system_vectors: SystemVectors, /, ) -> None:
        """
        Construct solution vectors and initial conditions indicator
        """
        self.transition_variables, self.initial_conditions = _solution_vector_from_system_vector(system_vectors.transition_variables, system_vectors.initial_conditions)
        self.transition_shocks = system_vectors.transition_shocks
        self.measurement_variables = system_vectors.measurement_variables
        self.measurement_shocks = system_vectors.measurement_shocks

    def get_initials(
        self,
        /, 
        kind: Literal["required"] | Literal["discarded"] = "required",
    ) -> Iterable[_incidence.Token]:
        """
        Get tokens representing required initial conditions
        """
        return list(it_.compress(self.transition_variables, self.initial_conditions))
    #]


def _create_system_transition_vector(tokens_transition_variables: Iterable[_incidence.Token], /, ) -> Iterable[_incidence.Token]:
    """
    From tokens of transition variables, create vector of transition variables
    along columns of matrix A in unsolved system
    """
    #[
    tokens_transition_variables = set(tokens_transition_variables)
    min_shifts = _incidence.get_some_shift_by_quantities(tokens_transition_variables, lambda x: min(min(x), -1))
    max_shifts = _incidence.get_some_shift_by_quantities(tokens_transition_variables, max)
    #
    vector_for_id = lambda qid: [_incidence.Token(qid, sh) for sh in range(min_shifts[qid]+1, max_shifts[qid]+1)]
    unique_ids = set(t.qid for t in tokens_transition_variables)
    return it_.chain.from_iterable(vector_for_id(i) for i in unique_ids)
    #]


def _solution_vector_from_system_vector(
    system_transition_vector: Iterable[_incidence.Token], 
    initial_conditions: Iterable[bool],
    /,
) -> Iterable[_incidence.Token]:
    """
    From sorted system vector, get vector of transition variables in solved
    system and the indicator of required initial conditions
    """
    num_forwards = _get_num_forwards(system_transition_vector)
    return system_transition_vector[num_forwards:], initial_conditions[num_forwards:]


def _get_num_forwards(system_transition_vector: Iterable[_incidence.Token]):
    """
    Number of forward-looking tokens in a vector of tokens
    """
    return sum(1 for t in system_transition_vector if t.shift>0)


def _get_num_backwards(system_transition_vector: Iterable[_incidence.Token]):
    """
    Number of backward-looking tokens in a vector of tokens
    """
    return len(system_transition_vector) - _get_num_forwards(system_transition_vector)


@_dc.dataclass(slots=True, )
class SystemMap:
    """
    """
    #[
    A: _maps.ArrayMap | None = None
    B: _maps.ArrayMap | None = None
    C: None = None
    D: _maps.ArrayMap | None = None
    dynid_A: np_.ndarray | None = None
    dynid_B: np_.ndarray | None = None
    dynid_C: np_.ndarray | None = None
    dynid_D: np_.ndarray | None = None
    #
    F: _maps.ArrayMap | None = None
    G: _maps.ArrayMap | None = None
    H: None = None
    J: _maps.ArrayMap | None = None

    def __init__(
        self,
        system_vectors: SystemVectors,
    ) -> None:
        """
        """
        system_eids = system_vectors.transition_eids + system_vectors.measurement_eids
        #
        # Create the map from equation ids to rhs offset; the offset is the
        # number of rows in the Jacobian matrix that precede the equation
        eid_to_rhs_offset = _maps.create_eid_to_rhs_offset(system_eids, system_vectors.eid_to_wrt_tokens)
        #
        # Transition equations
        #
        self.A = _maps.ArrayMap.for_equations(
            system_vectors.transition_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.transition_variables,
            eid_to_rhs_offset,
            #
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        lagged_transition_variables = [ t.shifted(-1) for t in system_vectors.transition_variables ]
        lagged_transition_variables = [ 
            t if t not in system_vectors.transition_variables else None 
            for t in lagged_transition_variables 
        ]
        #
        self.B = _maps.ArrayMap.for_equations(
            system_vectors.transition_eids,
            system_vectors.eid_to_wrt_tokens,
            lagged_transition_variables, 
            eid_to_rhs_offset,
            #
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        self.A.remove_nones()
        self.B.remove_nones()
        #
        self.C = _maps.ArrayMap.constant_vector(system_vectors.transition_eids)
        #
        self.D = _maps.ArrayMap.for_equations(
            system_vectors.transition_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.transition_shocks,
            eid_to_rhs_offset,
            #
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        num_dynid_rows = len(system_vectors.transition_variables) - len(system_vectors.transition_eids)
        self.dynid_A, self.dynid_B = _create_dynid_matrices(system_vectors.transition_variables, )
        self.dynid_C = np_.zeros((num_dynid_rows, system_vectors.shape_C_excl_dynid[1]), dtype=float, )
        self.dynid_D = np_.zeros((num_dynid_rows, system_vectors.shape_D_excl_dynid[1]), dtype=float, )
        #
        # Measurement equations
        #
        self.F = _maps.ArrayMap.for_equations(
            system_vectors.measurement_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.measurement_variables, 
            eid_to_rhs_offset,
            #
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        self.G = _maps.ArrayMap.for_equations(
            system_vectors.measurement_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.transition_variables, 
            eid_to_rhs_offset,
            #
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        self.H = _maps.ArrayMap.constant_vector(system_vectors.measurement_eids)
        #
        self.J = _maps.ArrayMap.for_equations(
            system_vectors.measurement_eids,
            system_vectors.eid_to_wrt_tokens,
            system_vectors.measurement_shocks, 
            eid_to_rhs_offset,
            #
            rhs_column=0,
            lhs_column_offset=0,
        )
    #]


def _create_dynid_matrices(system_transition_vector: Iterable[_incidence.Token]):
    """
    Create dynamic identity matrix for unsolved system
    """
    #[
    max_shifts = _incidence.get_some_shift_by_quantities(system_transition_vector, max)
    #
    index_A = ([], [])
    index_B = ([], [])
    row_count = 0
    for i, t in enumerate(system_transition_vector):
        if t.shift==max_shifts[t.qid]:
            continue
        j = system_transition_vector.index(t.shifted(+1))
        index_A[0].append(row_count)
        index_A[1].append(i)
        index_B[0].append(row_count)
        index_B[1].append(j)
        row_count += 1
    #
    num_columns = len(system_transition_vector)
    dynid_A = np_.zeros((row_count, num_columns), dtype=float)
    dynid_B = np_.zeros((row_count, num_columns), dtype=float)
    dynid_A[index_A] = 1
    dynid_B[index_B] = -1
    return dynid_A, dynid_B
    #]


def _adjust_for_measurement_equations(
    tokens_transition_variables: Iterable[_quantities.Quantity],
    equations: Iterable[_equations.Equation],
    qid_to_kind: dict[int, _quantities.QuantityKind],
    /,
) -> Iterable[_quantities.Quantity]:
    """
    """
    #[
    tokens_in_measurement_equations = it_.chain.from_iterable(e.incidence for e in equations if e.kind is _equations.EquationKind.MEASUREMENT_EQUATION)
    pretend_needed = [
        _incidence.Token(t.qid, t.shift-1) for t in tokens_in_measurement_equations
        if qid_to_kind[t.qid] in _quantities.QuantityKind.TRANSITION_VARIABLE
    ]
    return set(tokens_transition_variables).union(pretend_needed)
    #]


def _select_system_equations_from_equations(
    equations: Iterable[_equations.Equation],
    transition_eids: list[int],
    measurement_eids: list[int],
    /
) -> Iterable[_equations.Equation]:
    eid_to_equation = { eqn.id:eqn for eqn in equations }
    system_eids = transition_eids + measurement_eids
    return ( eid_to_equation[eid] for eid in system_eids )


